﻿/*
 Copyright (C) 2016 Gottfried Grosshans

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2 as published by the Free Software Foundation.
 */

#include <avr/sleep.h>
#include <SPI.h>
#include "nRF24L01.h"
#include "RF24.h"
#include "printf.h"

//
// Change the value of *channel* for another room!
//

byte const channel = 55;

byte const led = 7;
byte const sensor = 2;
byte const relaya = 5;
byte const relayb = 6;

bool again = false;

unsigned long angle = 999;

//
// Hardware configuration
//

// Set up nRF24L01 radio on SPI bus plus pins 9 & 10

RF24 radio(9,10);

//
// Topology
//

// Radio pipe addresses for the 2 nodes to communicate.
const uint64_t pipes[2] = { 0xF0F0F0F0E1LL, 0xF0F0F0F0D2LL };

//
// Role management
//
// Set up role.
//

// The various roles supported by this sketch
typedef enum { role_ping_out = 1, role_pong_back } role_e;

// The debug-friendly names of those roles
const char* role_friendly_name[] = { "invalid", "Ping out", "Pong back"};

// The role of the current running sketch
role_e role = role_pong_back;

// Interrupt handling routine

void wake ()
{
  // cancel sleep as a precaution
  sleep_disable();
  // precautionary while we do other stuff
  detachInterrupt (0);
}  // end of wake

// Turn hvac relay off

void hvacoff () {
    pinMode(led, OUTPUT);
    pinMode(relaya, OUTPUT);
    pinMode(relayb, OUTPUT);
    digitalWrite(led, HIGH);
    digitalWrite(relaya, HIGH);
    digitalWrite(relayb, LOW);
    delay(100);
    pinMode(led, INPUT);
    pinMode(relaya, INPUT);
    pinMode(relayb, INPUT);
  }

// Turn hvac relay on

void hvacon () {
    pinMode(led, OUTPUT);
    pinMode(relaya, OUTPUT);
    pinMode(relayb, OUTPUT);
    digitalWrite(led, HIGH);
    digitalWrite(relaya, LOW);
    digitalWrite(relayb, HIGH);
    delay(100);
    pinMode(led, INPUT);
    pinMode(relaya, INPUT);
    pinMode(relayb, INPUT);
  }

// switch from send to receive/nul/pong

void switchoff () {
     role = role_pong_back;
     radio.openWritingPipe(pipes[1]);
     radio.openReadingPipe(1,pipes[0]);
   }

void pingoutrole () {
  if (role == role_ping_out)
  {
// First, stop listening so we can talk.
    radio.stopListening();

// Take the time, and send it.  This will block until complete
    unsigned long time = millis();
//printf("Now sending %lu...",time);
//    bool ok = radio.write( &time, sizeof(unsigned long) );
//    printf("Will send %lu\n\r",angle);
    bool ok = radio.write( &angle, sizeof(unsigned long) );

// Now, continue listening
    radio.startListening();

// Wait here until we get a response, or timeout
    unsigned long started_waiting_at = millis();
    bool timeout = false;
    while ( ! radio.available() && ! timeout )
      if (millis() - started_waiting_at > 5 )
        timeout = true;

    if ( timeout )
    { // Do nul.
    }
    else
    {
// Grab the response, compare, and send to debugging spew
      radio.read( &angle, sizeof(unsigned long) );

// Spew it
      printf("Got response %lu, round-trip delay: %lu, Angle: %lu\n\r",time,millis()-time,angle);
      delay(100);

// Handle HVAC relay
      if (angle == 0) { hvacoff ();}
      if (angle == 255) { delay(900);  hvacon ();}
      angle=999;
      switchoff ();
      again = false;
    }

  }
}

void setup(void)
{
//
// Print preamble
//
  pinMode (sensor, INPUT);  // enable pull-up

  Serial.begin(57600);
  printf_begin();
  printf("\n\rHVAC CONTROL\n\r");
  printf("ROLE: %s\n\r",role_friendly_name[role]);
  printf("*** Connection:\n\r");

  //
  // Setup and configure rf radio
  //

  radio.begin();

  radio.setDataRate(RF24_250KBPS);
  radio.setChannel(channel);

  // optionally, increase the delay between retries & # of retries
  radio.setRetries(5,5);

  // optionally, reduce the payload size.  seems to
  // improve reliability
  radio.setPayloadSize(8);

  //
  // Open pipes to other nodes for communication
  //

  // This simple sketch opens two pipes for these two nodes to communicate
  // back and forth.
  // Open 'our' pipe for writing
  // Open the 'other' pipe for reading,
  // in position #1 (we can have up to 5 pipes open for reading)

  if ( role == role_ping_out )
  {
    radio.openWritingPipe(pipes[0]);
    radio.openReadingPipe(1,pipes[1]);
  }
  else
  {
    radio.openWritingPipe(pipes[1]);
    radio.openReadingPipe(1,pipes[0]);
  }

  //
  // Start listening
  //

  radio.startListening();

  //
  // Dump the configuration of the rf unit for debugging
  //

  radio.printDetails();
  radio.stopListening();
  radio.powerDown ();
}

// HVAC wakeup and control cycle

void wakeup() {
  radio.powerUp ();
  radio.startListening();
  delay(100);
  if (digitalRead(sensor) == HIGH) {angle = 0;}
  // tstat open = vent close / hvac off
  else {angle = 255;} // tstat closed = vent open / hvac on
  role = role_ping_out;
  radio.openWritingPipe(pipes[0]);
  radio.openReadingPipe(1,pipes[1]);
  again = true;
  while (again == true)
    {
  //
  // Ping out role.  Repeatedly send the current time
  //
    pingoutrole ();
  }
  radio.stopListening();
  radio.powerDown ();
}


void loop ()
{
  // disable ADC to save energy
  ADCSRA = 0;

  set_sleep_mode (SLEEP_MODE_PWR_DOWN);
  sleep_enable();

  // Do not interrupt before we go to sleep, or the
  // ISR will detach interrupts and we won't wake.
  noInterrupts ();

  // will be called when pin D2 goes low
  attachInterrupt (0, wake, CHANGE);
  EIFR = bit (INTF0);  // clear flag for interrupt 0

  // turn off brown-out enable in software
  // BODS must be set to one and BODSE must be set to zero within four clock cycles
  MCUCR = bit (BODS) | bit (BODSE);
  // The BODS bit is automatically cleared after three clock cycles
  MCUCR = bit (BODS);

  // We are guaranteed that the sleep_cpu call will be done
  // as the processor executes the next instruction after
  // interrupts are turned on.
  interrupts ();  // one cycle
  sleep_cpu ();   // one cycle

  wakeup ();

  } // end of loop delay(900);
